@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2011-2015, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -22,19 +22,86 @@ end
 
 local rules = {}
 local rspamd_logger = require "rspamd_logger"
-local cdb = require "rspamd_cdb"
-local util = require "rspamd_util"
-local regexp = require "rspamd_regexp"
+local rspamd_util = require "rspamd_util"
+local rspamd_regexp = require "rspamd_regexp"
 local rspamd_expression = require "rspamd_expression"
 local rspamd_ip = require "rspamd_ip"
 local lua_util = require "lua_util"
-local rspamd_dns = require "rspamd_dns"
 local lua_selectors = require "lua_selectors"
+local lua_maps = require "lua_maps"
 local redis_params
 local fun = require "fun"
 local N = 'multimap'
 
-local urls = {}
+local multimap_grammar
+-- Parse result in form: <symbol>:<score>|<symbol>|<score>
+local function parse_multimap_value(parse_rule, p_ret)
+  if p_ret and type(p_ret) == 'string' then
+    local lpeg = require "lpeg"
+
+    if not multimap_grammar then
+      local number = {}
+
+      local digit = lpeg.R("09")
+      number.integer =
+      (lpeg.S("+-") ^ -1) *
+          (digit   ^  1)
+
+      -- Matches: .6, .899, .9999873
+      number.fractional =
+      (lpeg.P(".")   ) *
+          (digit ^ 1)
+
+      -- Matches: 55.97, -90.8, .9
+      number.decimal =
+      (number.integer *              -- Integer
+          (number.fractional ^ -1)) +    -- Fractional
+          (lpeg.S("+-") * number.fractional)  -- Completely fractional number
+
+      local sym_start = lpeg.R("az", "AZ") + lpeg.S("_")
+      local sym_elt = sym_start + lpeg.R("09")
+      local symbol = sym_start * sym_elt ^0
+      local symbol_cap = lpeg.Cg(symbol, 'symbol')
+      local score_cap = lpeg.Cg(number.decimal, 'score')
+      local opts_cap = lpeg.Cg(lpeg.Ct(lpeg.C(symbol) * (lpeg.P(",") * lpeg.C(symbol))^0), 'opts')
+      local symscore_cap = (symbol_cap * lpeg.P(":") * score_cap)
+      local symscoreopt_cap = symscore_cap * lpeg.P(":") * opts_cap
+      local grammar = symscoreopt_cap + symscore_cap + symbol_cap + score_cap
+      multimap_grammar = lpeg.Ct(grammar)
+    end
+    local tbl = multimap_grammar:match(p_ret)
+
+    if tbl then
+      local sym
+      local score = 1.0
+      local opts = {}
+
+      if tbl.symbol then
+        sym = tbl.symbol
+      end
+      if tbl.score then
+        score = tonumber(tbl.score)
+      end
+      if tbl.opts then
+        opts = tbl.opts
+      end
+
+      return true,sym,score,opts
+    else
+      if p_ret ~= '' then
+        rspamd_logger.infox(rspamd_config, '%s: cannot parse string "%s"',
+            parse_rule.symbol, p_ret)
+      end
+
+      return true,nil,1.0,{}
+    end
+  elseif type(p_ret) == 'boolean' then
+    return p_ret,nil,1.0,{}
+  end
+
+  return false,nil,0.0,{}
+end
+
 
 local value_types = {
   ip = {
@@ -96,10 +163,10 @@ end
 
 local function apply_hostname_filter(task, filter, hostname, r)
   if filter == 'tld' then
-    local tld = util.get_tld(hostname)
+    local tld = rspamd_util.get_tld(hostname)
     return tld
   elseif filter == 'top' then
-    local tld = util.get_tld(hostname)
+    local tld = rspamd_util.get_tld(hostname)
     return tld:match('[^.]*$') or tld
   else
     if not r['re_filter'] then
@@ -108,13 +175,13 @@ local function apply_hostname_filter(task, filter, hostname, r)
         rspamd_logger.errx(task, 'bad search filter: %s', filter)
         return
       end
-      r['re_filter'] = regexp.create(pat)
+      r['re_filter'] = rspamd_regexp.create_cached(pat)
       if not r['re_filter'] then
         rspamd_logger.errx(task, 'couldnt create regex: %s', pat)
         return
       end
     end
-    local tld = util.get_tld(hostname)
+    local tld = rspamd_util.get_tld(hostname)
     local res = r['re_filter']:search(tld)
     if res then
       return res[1]
@@ -171,7 +238,7 @@ local function apply_url_filter(task, filter, url, r)
     if not r['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        r['re_filter'] = regexp.create(pat)
+        r['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -189,7 +256,7 @@ local function apply_url_filter(task, filter, url, r)
     if not r['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        r['re_filter'] = regexp.create(pat)
+        r['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -207,7 +274,7 @@ local function apply_url_filter(task, filter, url, r)
     if not r['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        r['re_filter'] = regexp.create(pat)
+        r['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -236,36 +303,42 @@ end
 
 local function apply_addr_filter(task, filter, input, rule)
   if filter == 'email:addr' or filter == 'email' then
-    local addr = util.parse_mail_address(input, task:get_mempool())
+    local addr = rspamd_util.parse_mail_address(input, task:get_mempool(), 1024)
     if addr and addr[1] then
-      return addr[1]['addr']
+      return fun.totable(fun.map(function(a) return a.addr end, addr))
     end
   elseif filter == 'email:user' then
-    local addr = util.parse_mail_address(input, task:get_mempool())
+    local addr = rspamd_util.parse_mail_address(input, task:get_mempool(), 1024)
     if addr and addr[1] then
-      return addr[1]['user']
+      return fun.totable(fun.map(function(a) return a.user end, addr))
     end
   elseif filter == 'email:domain' then
-    local addr = util.parse_mail_address(input, task:get_mempool())
+    local addr = rspamd_util.parse_mail_address(input, task:get_mempool(), 1024)
     if addr and addr[1] then
-      return addr[1]['domain']
+      return fun.totable(fun.map(function(a) return a.domain end, addr))
     end
   elseif filter == 'email:domain:tld' then
-    local addr = util.parse_mail_address(input, task:get_mempool())
+    local addr = rspamd_util.parse_mail_address(input, task:get_mempool(), 1024)
     if addr and addr[1] then
-      return util.get_tld(addr[1]['domain'])
+      return fun.totable(fun.map(function(a) return rspamd_util.get_tld(a.domain) end, addr))
     end
   elseif filter == 'email:name' then
-    local addr = util.parse_mail_address(input, task:get_mempool())
+    local addr = rspamd_util.parse_mail_address(input, task:get_mempool(), 1024)
     if addr and addr[1] then
-      return addr[1]['name']
+      return fun.totable(fun.map(function(a) return a.name end, addr))
+    end
+  elseif filter == 'ip_addr' then
+    local ip_addr = rspamd_ip.from_string(input)
+
+    if ip_addr and ip_addr:is_valid() then
+      return ip_addr
     end
   else
     -- regexp case
     if not rule['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        rule['re_filter'] = regexp.create(pat)
+        rule['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -288,7 +361,7 @@ local function apply_filename_filter(task, filter, fn, r)
     if not r['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        r['re_filter'] = regexp.create(pat)
+        r['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -312,7 +385,7 @@ local function apply_regexp_filter(task, filter, fn, r)
     if not r['re_filter'] then
       local type,pat = string.match(filter, '(regexp:)(.+)')
       if type and pat then
-        r['re_filter'] = regexp.create(pat)
+        r['re_filter'] = rspamd_regexp.create_cached(pat)
       end
     end
 
@@ -377,141 +450,163 @@ local multimap_filters = {
   --content = apply_content_filter, -- Content filters are special :(
 }
 
-local multimap_grammar
+local function multimap_query_redis(key, task, value, callback)
+  local cmd = 'HGET'
+  if type(value) == 'userdata' and value.class == 'rspamd{ip}' then
+    cmd = 'HMGET'
+  end
+
+  local srch = {key}
+
+  -- Insert all ips for some mask :(
+  if type(value) == 'userdata' and value.class == 'rspamd{ip}' then
+    srch[#srch + 1] = tostring(value)
+    -- IPv6 case
+    local maxbits = 128
+    local minbits = 64
+    if value:get_version() == 4 then
+      maxbits = 32
+      minbits = 8
+    end
+    for i=maxbits,minbits,-1 do
+      local nip = value:apply_mask(i):tostring() .. "/" .. i
+      srch[#srch + 1] = nip
+    end
+  else
+    srch[#srch + 1] = value
+  end
+
+  local function redis_map_cb(err, data)
+    lua_util.debugm(N, task, 'got reply from Redis when trying to get key %s: err=%s, data=%s',
+        key, err, data)
+    if not err and type(data) ~= 'userdata' then
+      callback(data)
+    end
+  end
+
+  return rspamd_redis_make_request(task,
+      redis_params, -- connect params
+      key, -- hash key
+      false, -- is write
+      redis_map_cb, --callback
+      cmd, -- command
+      srch -- arguments
+  )
+end
 
 local function multimap_callback(task, rule)
-  local pre_filter = rule['prefilter']
-
   local function match_element(r, value, callback)
-   if not value then
+    if not value then
       return false
     end
 
-    local function redis_map_cb(err, data)
-      if not err and type(data) ~= 'userdata' then
-        callback(data)
+    local function get_key_callback(ret, err_or_data, err_code)
+      lua_util.debugm(N, task, 'got return "%s" (err code = %s) for multimap %s',
+          err_or_data,
+          err_code,
+          rule.symbol)
+
+      if ret then
+        if type(err_or_data) == 'table' then
+          for _,elt in ipairs(err_or_data) do
+            callback(elt)
+          end
+        else
+          callback(err_or_data)
+        end
+      elseif err_code ~= 404 then
+        rspamd_logger.infox(task, "map %s: get key returned error %s: %s",
+          rule.symbol, err_code, err_or_data)
       end
     end
 
+    lua_util.debugm(N, task, 'check value %s for multimap %s', value,
+        rule.symbol)
+
     local ret = false
 
-    if r['cdb'] then
-      local srch = value
-      if r['type'] == 'ip' then
-        srch = value:to_string()
-      end
-      ret = r['cdb']:lookup(srch)
-    elseif r['redis_key'] then
-      local srch = {value}
-      local cmd = 'HGET'
-      if r['type'] == 'ip' or (r['type'] == 'received' and
-        (r['filter'] == 'real_ip' or r['filter'] == 'from_ip' or not r['filter'])) then
-        srch = {value:to_string()}
-        cmd = 'HMGET'
-        local maxbits = 128
-        local minbits = 32
-        if value:get_version() == 4 then
-            maxbits = 32
-            minbits = 8
-        end
-        for i=maxbits,minbits,-1 do
-            local nip = value:apply_mask(i):to_string() .. "/" .. i
-            table.insert(srch, nip)
-        end
-      end
+    if r.redis_key then
+      -- Deal with hash name here: it can be either plain string or a selector
+      if type(r.redis_key) == 'string' then
+        ret = multimap_query_redis(r.redis_key, task, value, callback)
+      else
+        -- Here we have a selector
+        local results = r.redis_key(task)
 
-      table.insert(srch, 1, r['redis_key'])
-      ret = rspamd_redis_make_request(task,
-        redis_params, -- connect params
-        r['redis_key'], -- hash key
-        false, -- is write
-        redis_map_cb, --callback
-        cmd, -- command
-        srch -- arguments
-      )
+        -- Here we need to spill this function into multiple queries
+        if type(results) == 'table' then
+          for _,res in ipairs(results) do
+            ret = multimap_query_redis(res, task, value, callback)
+
+            if not ret then
+              break
+            end
+          end
+        else
+          ret = multimap_query_redis(results, task, value, callback)
+        end
+      end
 
       return ret
-    elseif r['radix'] then
-      ret = r['radix']:get_key(value)
-    elseif r['hash'] then
-      ret = r['hash']:get_key(value)
+    elseif r.map_obj then
+      r.map_obj:get_key(value, get_key_callback, task)
     end
-
-    if ret then
-      if type(ret) == 'table' then
-        for _,elt in ipairs(ret) do
-          callback(elt)
-        end
-
-        ret = true
-      else
-        callback(ret)
-      end
-    end
-
-    return ret
   end
 
-  -- Parse result in form: <symbol>:<score>|<symbol>|<score>
-  local function parse_ret(parse_rule, p_ret)
-    if p_ret and type(p_ret) == 'string' then
-      local lpeg = require "lpeg"
-
-      if not multimap_grammar then
-      local number = {}
-
-        local digit = lpeg.R("09")
-        number.integer =
-        (lpeg.S("+-") ^ -1) *
-                (digit   ^  1)
-
-        -- Matches: .6, .899, .9999873
-        number.fractional =
-        (lpeg.P(".")   ) *
-                (digit ^ 1)
-
-        -- Matches: 55.97, -90.8, .9
-        number.decimal =
-        (number.integer *              -- Integer
-                (number.fractional ^ -1)) +    -- Fractional
-                (lpeg.S("+-") * number.fractional)  -- Completely fractional number
-
-        local sym_start = lpeg.R("az", "AZ") + lpeg.S("_")
-        local sym_elt = sym_start + lpeg.R("09")
-        local symbol = sym_start * sym_elt ^0
-        local symbol_cap = lpeg.Cg(symbol, 'symbol')
-        local score_cap = lpeg.Cg(number.decimal, 'score')
-        local symscore_cap = (symbol_cap * lpeg.P(":") * score_cap)
-        local grammar = symscore_cap + symbol_cap + score_cap
-        multimap_grammar = lpeg.Ct(grammar)
-      end
-      local tbl = multimap_grammar:match(p_ret)
-
-      if tbl then
-        local sym
-        local score = 1.0
-
-        if tbl['symbol'] then
-          sym = tbl['symbol']
+  local function insert_results(result, opt)
+    local _,symbol,score,opts = parse_multimap_value(rule, result)
+    local forced = false
+    if symbol then
+      if rule.symbols_set then
+        if not rule.symbols_set[symbol] then
+          rspamd_logger.infox(task, 'symbol %s is not registered for map %s, ' ..
+              'replace it with just %s',
+              symbol, rule.symbol, rule.symbol)
+          symbol = rule.symbol
         end
-        if tbl['score'] then
-          score = tbl['score']
+      elseif rule.disable_multisymbol then
+        symbol = rule.symbol
+        if type(opt) == 'table' then
+          table.insert(opt, result)
+        elseif type(opt) ~= nil then
+          opt = {opt,result}
+        else
+          opt = {result}
         end
-
-        return true,sym,score
       else
-        if p_ret ~= '' then
-          rspamd_logger.infox(task, '%s: cannot parse string "%s"',
-            parse_rule.symbol, p_ret)
-        end
-
-        return true,nil,1.0
+        forced = not rule.dynamic_symbols
       end
-    elseif type(p_ret) == 'boolean' then
-      return p_ret,nil,0.0
+    else
+      symbol = rule.symbol
     end
 
-    return false,nil,0.0
+    if opts and #opts > 0 then
+      -- Options come from the map itself
+      task:insert_result(forced, symbol, score, opts)
+    else
+      if opt then
+        if type(opt) == 'table' then
+          task:insert_result(forced, symbol, score, fun.totable(fun.map(tostring, opt)))
+        else
+          task:insert_result(forced, symbol, score, tostring(opt))
+        end
+
+      else
+        task:insert_result(forced, symbol, score)
+      end
+    end
+
+    if rule.action then
+      local message = rule.message
+      if rule.message_func then
+        message = rule.message_func(task, rule.symbol, opt)
+      end
+      if message then
+        task:set_pre_result(rule.action, message, N)
+      else
+        task:set_pre_result(rule.action, 'Matched map: ' .. rule.symbol, N)
+      end
+    end
   end
 
   -- Match a single value for against a single rule
@@ -526,52 +621,28 @@ local function multimap_callback(task, rule)
           end
           return
         end
-        local _,symbol,score = parse_ret(r, result)
-        local forced = false
-        if symbol then
-          if r['symbols_set'] then
-            if not r['symbols_set'][symbol] then
-              rspamd_logger.infox(task, 'symbol %s is not registered for map %s, ' ..
-                  'replace it with just %s',
-                  symbol, r['symbol'], r['symbol'])
-              symbol = r['symbol']
-            end
-          else
-            forced = true
-          end
-        else
-          symbol = r['symbol']
-        end
-
         local opt = value_types[r['type']].get_value(value)
-        if opt then
-          task:insert_result(forced, symbol, score, opt)
-        else
-          task:insert_result(forced, symbol, score)
-        end
-
-        if pre_filter then
-          if r['message_func'] then
-            r['message'] = r.message_func(task, r['symbol'], opt)
-          end
-          if r['message'] then
-            task:set_pre_result(r['action'], r['message'])
-          else
-            task:set_pre_result(r['action'], 'Matched map: ' .. r['symbol'])
-          end
-        end
+        insert_results(result, opt)
       end
     end
 
-    if r['filter'] or r['type'] == 'url' then
-      local fn = multimap_filters[r['type']]
+    if r.filter or r.type == 'url' then
+      local fn = multimap_filters[r.type]
 
       if fn then
-        value = fn(task, r['filter'], value, r)
+
+        local filtered_value = fn(task, r.filter, value, r)
+        lua_util.debugm(N, task, 'apply filter %s for rule %s: %s -> %s',
+            r.filter, r.symbol, value, filtered_value)
+        value = filtered_value
       end
     end
 
-    match_element(r, value, rule_callback)
+    if type(value) == 'table' then
+      fun.each(function(elt) match_element(r, elt, rule_callback) end, value)
+    else
+      match_element(r, value, rule_callback)
+    end
   end
 
   -- Match list of values according to the field
@@ -607,7 +678,7 @@ local function multimap_callback(task, rule)
   end
 
   local function match_hostname(r, hostname)
-     match_rule(r, hostname)
+    match_rule(r, hostname)
   end
 
   local function match_filename(r, fn)
@@ -675,7 +746,7 @@ local function multimap_callback(task, rule)
         end
       else
         if use_tld and type(v) == 'string' then
-          v = util.get_tld(v)
+          v = rspamd_util.get_tld(v)
         end
         match_rule(r, v)
       end
@@ -696,46 +767,55 @@ local function multimap_callback(task, rule)
     end
   end
 
-  if rule['expression'] then
+  if rule.expression and not rule.combined then
     local res,trace = rule['expression']:process_traced(task)
 
     if not res or res == 0 then
-      lua_util.debugm(N, task, 'condition is false for %s', rule['symbol'])
+      lua_util.debugm(N, task, 'condition is false for %s',
+          rule.symbol)
       return
     else
-      lua_util.debugm(N, task, 'condition is true for %s: %s', rule['symbol'],
-        trace)
+      lua_util.debugm(N, task, 'condition is true for %s: %s',
+          rule.symbol,
+          trace)
     end
   end
 
-  local rt = rule['type']
   local process_rule_funcs = {
+    ip = function()
+      local ip = task:get_from_ip()
+      if ip and ip:is_valid() then
+        match_rule(rule, ip)
+      end
+    end,
     dnsbl = function()
       local ip = task:get_from_ip()
-      if ip:is_valid() then
-        if rt == 'ip' then
-          match_rule(rule, ip)
-        else
-          local to_resolve = ip_to_rbl(ip, rule['map'])
+      if ip and ip:is_valid() then
+        local to_resolve = ip_to_rbl(ip, rule['map'])
+        local function dns_cb(_, _, results, err)
+          lua_util.debugm(N, rspamd_config,
+              'resolve() finished: results=%1, err=%2, to_resolve=%3',
+              results, err, to_resolve)
 
-          local is_ok, results = rspamd_dns.request({
-            type = "a",
-            task = task,
-            name = to_resolve,
-          })
-
-          lua_util.debugm(N, rspamd_config, 'resolve() finished: results=%1, is_ok=%2, to_resolve=%3', results, is_ok, to_resolve)
-
-          if not is_ok and (results ~= 'requested record is not found' and results ~= 'no records with this name') then
+          if err and
+              (err ~= 'requested record is not found' and
+                  err ~= 'no records with this name') then
             rspamd_logger.errx(task, 'error looking up %s: %s', to_resolve, results)
-          elseif is_ok then
+          elseif results then
             task:insert_result(rule['symbol'], 1, rule['map'])
-            if pre_filter then
-              task:set_pre_result(rule['action'], 'Matched map: ' .. rule['symbol'])
+            if rule.action then
+              task:set_pre_result(rule['action'],
+                  'Matched map: ' .. rule['symbol'], N)
             end
           end
-
         end
+
+        task:get_resolver():resolve_a({
+          task= task,
+          name = to_resolve,
+          callback = dns_cb,
+          forced = true
+        })
       end
     end,
     header = function()
@@ -782,11 +862,38 @@ local function multimap_callback(task, rule)
         end
       end
     end,
+    user = function()
+      local user = task:get_user()
+      if user then
+        match_rule(rule, user)
+      end
+    end,
     filename = function()
       local parts = task:get_parts()
-      for _,p in ipairs(parts) do
-        if p:is_archive() and not rule['skip_archives'] then
-          local fnames = p:get_archive():get_files()
+
+      local function filter_parts(p)
+        return p:is_attachment() or (not p:is_text()) and (not p:is_multipart())
+      end
+
+      local function filter_archive(p)
+        local ext = p:get_detected_ext()
+        local det_type = 'unknown'
+
+        if ext then
+          local lua_magic_types = require "lua_magic/types"
+          local det_t = lua_magic_types[ext]
+
+          if det_t then
+            det_type = det_t.type
+          end
+        end
+
+        return p:is_archive() and det_type == 'archive' and not rule.skip_archives
+      end
+
+      for _,p in fun.iter(fun.filter(filter_parts, parts)) do
+        if filter_archive(p) then
+          local fnames = p:get_archive():get_files(1000)
 
           for _,fn in ipairs(fnames) do
             match_filename(rule, fn)
@@ -797,8 +904,20 @@ local function multimap_callback(task, rule)
         if fn then
           match_filename(rule, fn)
         end
+        -- Also deal with detected content type
+        if not rule.skip_detected then
+          local ext = p:get_detected_ext()
+
+          if ext then
+            local fake_fname = string.format('detected.%s', ext)
+            lua_util.debugm(N, task, 'detected filename %s',
+                fake_fname)
+            match_filename(rule, fake_fname)
+          end
+        end
       end
     end,
+
     content = function()
       match_content(rule)
     end,
@@ -860,71 +979,112 @@ local function multimap_callback(task, rule)
         end
       end
     end,
+    combined = function()
+      local ret,trace = rule.combined:process(task)
+      if ret and ret ~= 0 then
+        for n,t in pairs(trace) do
+          insert_results(t.value, string.format("%s=%s",
+              n, t.matched))
+        end
+      end
+    end,
   }
 
-  process_rule_funcs.ip = process_rule_funcs.dnsbl
-  local f = process_rule_funcs[rt]
-  if f then
-    f()
+  local rt = rule.type
+  local process_func = process_rule_funcs[rt]
+  if process_func then
+    process_func()
   else
     rspamd_logger.errx(task, 'Unrecognised rule type: %s', rt)
   end
 end
 
+
 local function gen_multimap_callback(rule)
   return function(task)
     multimap_callback(task, rule)
   end
 end
 
+local function multimap_on_load_gen(rule)
+  return function()
+    lua_util.debugm(N, rspamd_config, "loaded map object for rule %s", rule['symbol'])
+    local known_symbols = {}
+    rule.map_obj:foreach(function(key, value)
+      local r,symbol,score,_ = parse_multimap_value(rule, value)
+
+      if r and symbol and not known_symbols[symbol] then
+        lua_util.debugm(N, rspamd_config, "%s: adding new symbol %s (score = %s), triggered by %s",
+            rule.symbol, symbol, score, key)
+        rspamd_config:register_symbol{
+          name = value,
+          parent = rule.callback_id,
+          type = 'virtual',
+          score = score,
+        }
+        rspamd_config:set_metric_symbol({
+          group = N,
+          score = 1.0, -- In future, we will parse score from `get_value` and use it as multiplier
+          description = 'Automatic symbol generated by rule: ' .. rule.symbol,
+          name = value,
+        })
+        known_symbols[value] = true
+      end
+    end)
+  end
+end
+
 local function add_multimap_rule(key, newrule)
   local ret = false
 
-  local function multimap_load_hash(rule)
+  local function multimap_load_kv_map(rule)
     if rule['regexp'] then
       if rule['multi'] then
-        rule['hash'] = rspamd_config:add_map ({
-          url = rule['map'],
-          description = rule['description'],
-          type = 'regexp_multi'
-        })
+        rule.map_obj = lua_maps.map_add_from_ucl(rule.map, 'regexp_multi',
+            rule.description)
       else
-        rule['hash'] = rspamd_config:add_map ({
-          url = newrule['map'],
-          description = newrule['description'],
-          type = 'regexp'
-        })
+        rule.map_obj = lua_maps.map_add_from_ucl(rule.map, 'regexp',
+            rule.description)
       end
     elseif rule['glob'] then
       if rule['multi'] then
-        rule['hash'] = rspamd_config:add_map ({
-          url = rule['map'],
-          description = rule['description'],
-          type = 'glob_multi'
-        })
+        rule.map_obj = lua_maps.map_add_from_ucl(rule.map, 'glob_multi',
+            rule.description)
       else
-        rule['hash'] = rspamd_config:add_map ({
-          url = rule['map'],
-          description = rule['description'],
-          type = 'glob'
-        })
+        rule.map_obj = lua_maps.map_add_from_ucl(rule.map, 'glob',
+            rule.description)
       end
     else
-      rule['hash'] = rspamd_config:add_map ({
-        url = rule['map'],
-        description = rule['description'],
-        type = 'hash'
-      })
+      rule.map_obj = lua_maps.map_add_from_ucl(rule.map, 'hash',
+          rule.description)
     end
   end
 
+  local known_generic_types = {
+    header = true,
+    rcpt = true,
+    from = true,
+    helo = true,
+    symbol_options = true,
+    filename = true,
+    url = true,
+    user = true,
+    content = true,
+    hostname = true,
+    asn = true,
+    country = true,
+    mempool = true,
+    selector = true,
+    combined = true
+  }
+
   if newrule['message_func'] then
     newrule['message_func'] = assert(load(newrule['message_func']))()
   end
   if newrule['url'] and not newrule['map'] then
     newrule['map'] = newrule['url']
   end
-  if not newrule['map'] then
+  if not (newrule.map or newrule.rules)  then
     rspamd_logger.errx(rspamd_config, 'incomplete rule, missing map')
     return nil
   end
@@ -936,7 +1096,7 @@ local function add_multimap_rule(key, newrule)
   end
   if not newrule['description'] then
     newrule['description'] = string.format('multimap, type %s: %s', newrule['type'],
-      newrule['symbol'])
+        newrule['symbol'])
   end
   if newrule['type'] == 'mempool' and not newrule['variable'] then
     rspamd_logger.errx(rspamd_config, 'mempool map requires variable')
@@ -959,19 +1119,11 @@ local function add_multimap_rule(key, newrule)
       newrule.selector = selector
     end
   end
-  -- Check cdb flag
-  if type(newrule['map']) == 'string' and string.find(newrule['map'], '^cdb://.*$') then
-    newrule['cdb'] = cdb.create(newrule['map'])
-    if newrule['cdb'] then
-      ret = true
-    else
-      rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
-          newrule['map'])
-    end
-  elseif type(newrule['map']) == 'string' and string.find(newrule['map'], '^redis://.*$') then
+  if type(newrule['map']) == 'string' and
+      string.find(newrule['map'], '^redis://.*$') then
     if not redis_params then
       rspamd_logger.infox(rspamd_config, 'no redis servers are specified, ' ..
-        'cannot add redis map %s: %s', newrule['symbol'], newrule['map'])
+          'cannot add redis map %s: %s', newrule['symbol'], newrule['map'])
       return nil
     end
 
@@ -980,130 +1132,113 @@ local function add_multimap_rule(key, newrule)
     if newrule['redis_key'] then
       ret = true
     end
-  else
-    if type(newrule['map']) == 'string' then
-      local map = urls[newrule['map']]
-      if map and map['regexp'] == newrule['regexp'] and
-          map['glob'] == newrule['glob'] then
-        if newrule['type'] == 'ip' then
-          newrule['radix'] = map['map']
-        else
-          newrule['hash'] = map['map']
-        end
-        rspamd_logger.infox(rspamd_config, 'reuse url for %s: "%s"',
-          newrule['symbol'], newrule['map'])
-        ret = true
-      end
+  elseif type(newrule['map']) == 'string' and
+      string.find(newrule['map'], '^redis%+selector://.*$') then
+    if not redis_params then
+      rspamd_logger.infox(rspamd_config, 'no redis servers are specified, ' ..
+          'cannot add redis map %s: %s', newrule['symbol'], newrule['map'])
+      return nil
     end
-    if not ret then
-      if newrule['type'] == 'ip' then
-        newrule['radix'] = rspamd_config:add_map ({
-          url = newrule['map'],
-          description = newrule['description'],
-          type = 'radix'
-        })
-        if newrule['radix'] then
+
+    local selector_str = string.match(newrule['map'], '^redis%+selector://(.*)$')
+    local selector = lua_selectors.create_selector_closure(
+        rspamd_config, selector_str, newrule['delimiter'] or "")
+
+    if not selector then
+      rspamd_logger.errx(rspamd_config, 'redis selector map has invalid selector: "%s", symbol: %s',
+          selector_str, newrule['symbol'])
+      return nil
+    end
+
+    newrule['redis_key'] = selector
+    ret = true
+  elseif newrule.type == 'combined' then
+    local lua_maps_expressions = require "lua_maps_expressions"
+    newrule.combined = lua_maps_expressions.create(rspamd_config,
+        {
+          rules = newrule.rules,
+          expression = newrule.expression,
+          on_load = newrule.dynamic_symbols and multimap_on_load_gen(newrule) or nil,
+        }, N, 'Combined map for ' .. newrule.symbol)
+    if not newrule.combined then
+      rspamd_logger.errx(rspamd_config, 'cannot add combined map for %s', newrule.symbol)
+    else
+      ret = true
+    end
+  else
+    if newrule['type'] == 'ip' then
+      newrule.map_obj = lua_maps.map_add_from_ucl(newrule.map, 'radix',
+          newrule.description)
+      if newrule.map_obj then
+        ret = true
+      else
+        rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
+            newrule['map'])
+      end
+    elseif newrule['type'] == 'received' then
+      if type(newrule['flags']) == 'table' and newrule['flags'][1] then
+        newrule['flags'] = newrule['flags']
+      elseif type(newrule['flags']) == 'string' then
+        newrule['flags'] = {newrule['flags']}
+      end
+      if type(newrule['nflags']) == 'table' and newrule['nflags'][1] then
+        newrule['nflags'] = newrule['nflags']
+      elseif type(newrule['nflags']) == 'string' then
+        newrule['nflags'] = {newrule['nflags']}
+      end
+      local filter = newrule['filter'] or 'real_ip'
+      if filter == 'real_ip' or filter == 'from_ip' then
+        newrule.map_obj = lua_maps.map_add_from_ucl(newrule.map, 'radix',
+            newrule.description)
+        if newrule.map_obj then
           ret = true
-          if type(newrule['map']) == 'string' then
-            urls[newrule['map']] = {
-              type = 'ip',
-              map = newrule['radix'],
-              regexp = false
-            }
-          end
         else
           rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
-            newrule['map'])
-        end
-      elseif newrule['type'] == 'received' then
-        if type(newrule['flags']) == 'table' and newrule['flags'][1] then
-          newrule['flags'] = newrule['flags']
-        elseif type(newrule['flags']) == 'string' then
-          newrule['flags'] = {newrule['flags']}
-        end
-        if type(newrule['nflags']) == 'table' and newrule['nflags'][1] then
-          newrule['nflags'] = newrule['nflags']
-        elseif type(newrule['nflags']) == 'string' then
-          newrule['nflags'] = {newrule['nflags']}
-        end
-        local filter = newrule['filter'] or 'real_ip'
-        if filter == 'real_ip' or filter == 'from_ip' then
-          newrule['radix'] = rspamd_config:add_map ({
-            url = newrule['map'],
-            description = newrule['description'],
-            type = 'radix'
-          })
-          if newrule['radix'] then
-            ret = true
-          end
-        else
-          multimap_load_hash(newrule)
-
-          if newrule['hash'] then
-            ret = true
-            if type(newrule['map']) == 'string' then
-              urls[newrule['map']] = {
-                type = newrule['type'],
-                map = newrule['hash'],
-                regexp = newrule['regexp']
-              }
-            end
-          else
-            rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
               newrule['map'])
-          end
         end
-      elseif newrule['type'] == 'header'
-          or newrule['type'] == 'rcpt'
-          or newrule['type'] == 'from'
-          or newrule['type'] == 'helo'
-          or newrule['type'] == 'symbol_options'
-          or newrule['type'] == 'filename'
-          or newrule['type'] == 'url'
-          or newrule['type'] == 'content'
-          or newrule['type'] == 'hostname'
-          or newrule['type'] == 'asn'
-          or newrule['type'] == 'country'
-          or newrule['type'] == 'mempool'
-          or newrule['type'] == 'selector'then
+      else
+        multimap_load_kv_map(newrule)
 
-        multimap_load_hash(newrule)
-
-        if newrule['hash'] then
+        if newrule.map_obj then
           ret = true
-          if type(newrule['map']) == 'string' then
-            urls[newrule['map']] = {
-              type = newrule['type'],
-              map = newrule['hash'],
-              regexp = newrule['regexp']
-            }
-          end
         else
           rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
-            newrule['map'])
+              newrule['map'])
         end
-      elseif newrule['type'] == 'dnsbl' then
-        ret = true
       end
-    end
-  end
+    elseif known_generic_types[newrule.type] then
 
-  if newrule['action'] then
-    newrule['prefilter'] = true
-  else
-    newrule['prefilter'] = false
+      if newrule.filter == 'ip_addr' then
+        newrule.map_obj = lua_maps.map_add_from_ucl(newrule.map, 'radix',
+            newrule.description)
+      elseif not newrule.combined then
+        multimap_load_kv_map(newrule)
+      end
+
+      if newrule.map_obj then
+        ret = true
+      else
+        rspamd_logger.warnx(rspamd_config, 'Cannot add rule: map doesn\'t exists: %1',
+            newrule['map'])
+      end
+    elseif newrule['type'] == 'dnsbl' then
+      ret = true
+    end
   end
 
   if ret then
+    if newrule.map_obj and newrule.dynamic_symbols then
+      newrule.map_obj:on_load(multimap_on_load_gen(newrule))
+    end
     if newrule['type'] == 'symbol_options' then
       rspamd_config:register_dependency(newrule['symbol'], newrule['target_symbol'])
     end
-    if newrule['require_symbols'] and not newrule['prefilter'] then
+    if newrule['require_symbols'] then
       local atoms = {}
 
       local function parse_atom(str)
         local atom = table.concat(fun.totable(fun.take_while(function(c)
-          if string.find(', \t()><+!|&\n', c) then
+          if string.find(', \t()><+!|&\n', c, 1, true) then
             return false
           end
           return true
@@ -1124,13 +1259,13 @@ local function add_multimap_rule(key, newrule)
       end
 
       local expression = rspamd_expression.create(newrule['require_symbols'],
-        {parse_atom, process_atom}, rspamd_config:get_mempool())
+          {parse_atom, process_atom}, rspamd_config:get_mempool())
       if expression then
         newrule['expression'] = expression
 
         fun.each(function(v)
           lua_util.debugm(N, rspamd_config, 'add dependency %s -> %s',
-            newrule['symbol'], v)
+              newrule['symbol'], v)
           rspamd_config:register_dependency(newrule['symbol'], v)
         end, atoms)
       end
@@ -1159,11 +1294,21 @@ if opts and type(opts) == 'table' then
   end
   -- add fake symbol to check all maps inside a single callback
   fun.each(function(rule)
+    local augmentations = {}
+
+    if rule.action then
+      table.insert(augmentations, 'passthrough')
+    end
+
     local id = rspamd_config:register_symbol({
       type = 'normal',
       name = rule['symbol'],
+      augmentations = augmentations,
       callback = gen_multimap_callback(rule),
     })
+
+    rule.callback_id = id
+
     if rule['symbols'] then
       -- Find allowed symbols by this map
       rule['symbols_set'] = {}
@@ -1171,30 +1316,39 @@ if opts and type(opts) == 'table' then
         rspamd_config:register_symbol({
           type = 'virtual',
           name = s,
-          parent = id
+          parent = id,
+          score = tonumber(rule.score or "0") or 0, -- Default score
         })
         rule['symbols_set'][s] = 1
       end, rule['symbols'])
     end
-    if rule['score'] then
+    if not rule.score then
+      rspamd_logger.infox(rspamd_config, 'set default score 0 for multimap rule %s', rule.symbol)
+      rule.score = 0
+    end
+    if rule.score then
       -- Register metric symbol
       rule.name = rule.symbol
       rule.description = rule.description or 'multimap symbol'
       rule.group = rule.group or N
 
-      rspamd_config:set_metric_symbol(rule)
-    end
-  end,
-  fun.filter(function(r) return not r['prefilter'] end, rules))
+      local tmp_flags
+      tmp_flags = rule.flags
 
-  fun.each(function(r)
-    rspamd_config:register_symbol({
-      type = 'prefilter',
-      name = r['symbol'],
-      callback = gen_multimap_callback(r),
-    })
-  end,
-  fun.filter(function(r) return r['prefilter'] end, rules))
+      if rule.type == 'received' and rule.flags then
+        -- XXX: hack to allow received flags/nflags
+        -- See issue #3526 on GH
+        rule.flags = nil
+      end
+
+      -- XXX: for combined maps we use trace, so flags must include one_shot to avoid scores multiplication
+      if rule.combined and not rule.flags then
+        rule.flags = 'one_shot'
+      end
+      rspamd_config:set_metric_symbol(rule)
+      rule.flags = tmp_flags
+    end
+  end, rules)
 
   if #rules == 0 then
     lua_util.disable_module(N, "config")
